header {* Paths and Other Basic Definitions *}

theory Basics imports Main begin


(* Notation *)

declare [[coercion_enabled]]

notation Some ("_\<^sub>!" [1000] 1000)


(* Basic (Infinite) Token Types *)

datatype label = Label string
datatype key   = Key   string
datatype token = Token string

theorem string_infinite [simp]:
  shows "\<not> finite (UNIV :: string set)"
  by (rule infinite_UNIV_listI)

theorem label_inj [simp]:
  shows "inj Label"
  unfolding inj_on_def by simp

theorem key_inj [simp]:
  shows "inj Key"
  unfolding inj_on_def by simp

theorem token_inj [simp]:
  shows "inj Token"
  unfolding inj_on_def by simp

theorem label_infinite [simp]:
  shows "\<not> finite (UNIV :: label set)"
proof
  assume "finite (UNIV :: label set)"
  hence "finite (UNIV :: string set)" by - (drule finite_vimageI[of _ Label], auto)
  thus False by simp 
qed

theorem key_infinite [simp]:
  shows "\<not> finite (UNIV :: key set)"
proof
  assume "finite (UNIV :: key set)"
  hence "finite (UNIV :: string set)" by - (drule finite_vimageI[of _ Key], auto)
  thus False by simp 
qed

theorem token_infinite [simp]:
  shows "\<not> finite (UNIV :: token set)"
proof
  assume "finite (UNIV :: token set)"
  hence "finite (UNIV :: string set)" by - (drule finite_vimageI[of _ Token], auto)
  thus False by simp 
qed


corollary fresh_label [dest]:
  fixes m :: "label set" assumes "finite m"
  shows "\<exists>l. l \<notin> m"
  using assms by - (erule ex_new_if_finite[rotated], auto)

corollary fresh_key [dest]:
  fixes m :: "key set" assumes "finite m"
  shows "\<exists>l. l \<notin> m"
  using assms by - (erule ex_new_if_finite[rotated], auto)

corollary fresh_token:
  fixes m :: "token set" assumes "finite m"
  shows "\<exists>l. l \<notin> m"
  using assms by - (erule ex_new_if_finite[rotated], auto)


definition null :: key where
  "null \<equiv> Key Nil"

abbreviation constNull :: "'a \<Rightarrow> key" ("c\<^sub>N") where
  "c\<^sub>N \<equiv> \<lambda>x. null"


(* Paths *)

no_notation divide (infixl "'/" 70)

datatype path = 
  Root
| Child path label key ("_'/_[_]" [1000, 0, 0] 1000)

theorem path_child [elim_format, elim!]:
  assumes "p \<noteq> Root" and "\<And>q l k. p = q/l[k] \<Longrightarrow> P"
  shows P
  using assms by (cases p, auto)


(* Kinds *)

type_synonym kind = "label list"

abbreviation EmptyKind :: "kind" where
  "EmptyKind \<equiv> []"

abbreviation Step :: "label \<Rightarrow> kind \<Rightarrow> kind" where
  "Step l k \<equiv> k @ [l]"


class kindOf =
fixes kindOf :: "'a \<rightharpoonup> kind"

instantiation path :: kindOf begin 

  primrec kindOf_path where
    "kindOf_path Root   = EmptyKind\<^sub>!"
  | "kindOf_path p/l[i] = Option.map (Step l) (kindOf_path p)"

instance by intro_classes end


theorem kind_path:
  fixes p :: path
  shows  "\<exists>k. kindOf p = k\<^sub>!"
  by (induct p) auto

corollary kind_path_not_none [simp]:
  fixes p :: path
  shows "kindOf p = None \<longleftrightarrow> False"
  by (induct p) auto

theorem kind_path_option_map [elim_format, elim!]:
  fixes p :: path
  assumes "kindOf p = Option.map (Step l) r"
  shows "\<exists>k. kindOf p = (Step l k)\<^sub>! \<and> r = k\<^sub>!" 
  using assms by auto (case_tac r, auto)


abbreviation kind :: "kind \<Rightarrow> ('a :: kindOf) set" where 
  "kind k \<equiv> {p. kindOf p = k\<^sub>!}"

theorem kind_empty_only_root [simp]:
  shows "kind EmptyKind = {Root}"
  by auto (case_tac x, auto)


abbreviation hasKind where
  "hasKind k s \<equiv> \<forall>x \<in> s. kindOf x = k\<^sub>!"


(* Parents, Ancestors & Descendants *)

primrec parent :: "path \<Rightarrow> path" ("_'/.." [1000] 1000) where
  "Root/.. = Root"
| "p/l[i]/.. = p"

fun anc :: "path \<Rightarrow> path set" ("anc _" [1000] 100) where
  "anc Root = {Root}"
| "anc p = insert p (anc p/..)"

abbreviation desc :: "path \<Rightarrow> path set" ("desc _" [1000] 100) where
  "desc q \<equiv> {p. q \<in> anc p}"

theorem parent_kind [simp]:
  assumes "p \<noteq> Root"
  shows "kindOf p/.. \<noteq> kindOf p"
  using assms by (cases p) auto

theorem anc_reflex [simp]:
  shows "p \<in> anc p"
  by (cases p) auto

theorem anc_root [simp]:
  shows "Root \<in> anc p"
  by (induct p) auto

theorem anc_child_right [simp]:
  assumes "p \<in> anc q"
  shows "p \<in> anc q/l[i]"
  using assms by (induct q) auto

theorem anc_child_left [intro]:
  assumes "p/l[i] \<in> anc q"
  shows "p \<in> anc q"
  using assms by (induct q) auto

theorem anc_parent_left [intro]:
  assumes "p \<in> anc q/.."
  shows "p \<in> anc q"
  using assms by (induct q) auto

theorem anc_parent_right [simp]:
  assumes "p \<in> anc q"
  shows "p/.. \<in> anc q"
  using assms by (induct q) auto

theorem anc_child_not [simp]:
  shows "p/l[i] \<notin> anc p"
  by (induct p arbitrary: l i, auto) (metis anc_reflex)

theorem anc_not_parent [simp]:
  assumes "p \<noteq> Root"
  shows "p \<notin> anc p/.."
  using assms by (cases p, auto)

theorem anc_kind [simp]:
  assumes "p \<in> anc q"
  shows "kindOf p = kindOf q \<longleftrightarrow> p = q"
  using assms by -
   (induct q arbitrary: p, auto, case_tac p, auto, metis anc_child_left anc_child_not)

theorem anc_kind_extend:
  assumes "p \<in> anc q" and "kindOf p = k\<^sub>!"
  shows "\<exists>s. kindOf q = (k @ s)\<^sub>!"
  using assms by - (induct q rule: anc.induct, auto)

theorem desc_kind [simp]:
  assumes "p \<in> desc q"
  shows "kindOf p = kindOf q \<longleftrightarrow> p = q"
  using assms by auto (metis anc_kind)

theorem desc_kind_extend:
  assumes "p \<in> desc q" and "kindOf q = k\<^sub>!"
  shows "\<exists>s. kindOf p = (k @ s)\<^sub>!"
  using assms by auto (metis anc_kind_extend)


end
