header {* Documents and Values *}

theory Documents imports Basics begin


(* Basic Values & Types *)

datatype "value" =
  VKey key     ("_\<^sub>k" [1000] 1000)
| VInt int     ("_\<^sub>i" [1000] 1000)
| VToken token ("_\<^sub>t" [1000] 1000)
| Unit

declare [[coercion VInt]]
declare [[coercion VToken]]
declare [[coercion VKey]]

datatype type =
  TKey
| TInt
| TToken
| TUnit

primrec typeOf :: "value \<Rightarrow> type" where
  "typeOf (VKey _)   = TKey"
| "typeOf (VInt _)   = TInt"
| "typeOf (VToken _) = TToken"
| "typeOf Unit       = TUnit"


(* Discriminators & Selectors *)

abbreviation valuesOf :: "type \<Rightarrow> value set" where
  "valuesOf t \<equiv> {v. typeOf v = t}"

abbreviation keys :: "value set" where
  "keys \<equiv> valuesOf TKey"

abbreviation tokens :: "value set" where
  "tokens \<equiv> valuesOf TToken"

abbreviation ints :: "value set" where
  "ints \<equiv> valuesOf TInt"


fun asKey :: "value \<rightharpoonup> key" where
  "asKey i = i\<^sub>!"
| "asKey _ = None"

fun asInt :: "value \<rightharpoonup> int" where
  "asInt n = n\<^sub>!"
| "asInt _ = None"

fun asToken :: "value \<rightharpoonup> token" where
  "asToken s = s\<^sub>!"
| "asToken _ = None"


fun theKey :: "value \<Rightarrow> key" where
  "theKey i = i"
| "theKey _ = undefined"

fun theInt :: "value \<Rightarrow> int" where
  "theInt n = n"
| "theInt _ = undefined"

fun theToken :: "value \<Rightarrow> token" where
  "theToken s = s"
| "theToken _ = undefined"


(* Theorems *)

theorem vkey_inj [simp]:
  shows "inj VKey"
  unfolding inj_on_def by simp

lemma VKey_infinite_range [simp]:
  shows "\<not> finite (range VKey)"
  by clarify (metis finite_imageD key_infinite vkey_inj)

corollary value_infinite [simp]:
  shows "\<not> finite (UNIV :: value set)"
  by clarify (drule rev_finite_subset[of _ "range VKey"], auto)


theorem typeOf_not_int [simp]:
  shows "(\<forall>x. v \<noteq> x\<^sub>i) \<longleftrightarrow> typeOf v \<noteq> TInt"
  by (rule_tac y=v in value.exhaust) auto

theorem typeOf_not_str [simp]:
  shows "(\<forall>x. v \<noteq> x\<^sub>t) \<longleftrightarrow> typeOf v \<noteq> TToken"
  by (rule_tac y=v in value.exhaust) auto

theorem typeOf_not_id [simp]:
  shows "(\<forall>x. v \<noteq> x\<^sub>k) \<longleftrightarrow> typeOf v \<noteq> TKey"
  by (rule_tac y=v in value.exhaust) auto

theorem typeOf_TKeyD [dest!]:
  assumes "typeOf v = TKey"
  shows "\<exists>x. v = x\<^sub>k"
  using assms by (cases v) simp+

theorem typeOf_TIntD [dest!]:
  assumes "typeOf v = TInt"
  shows "\<exists>x. v = x\<^sub>i"
  using assms by (cases v) simp+

theorem typeOf_TTokenD [dest!]:
  assumes "typeOf v = TToken"
  shows "\<exists>x. v = x\<^sub>t"
  using assms by (cases v) simp+

theorem typeOf_TUnitD [dest!]:
  assumes "typeOf v = TUnit"
  shows "v = Unit"
  using assms by (cases v) simp+


theorem asKey_cast [simp]:
  shows "asKey x = y\<^sub>! \<longleftrightarrow> x = y"
  by (cases x) simp+

theorem asInt_cast [simp]:
  shows "asInt x = y\<^sub>! \<longleftrightarrow> x = y"
  by (cases x) simp+

theorem asToken_cast [simp]:
  shows "asToken x = y\<^sub>! \<longleftrightarrow> x = y"
  by (cases x) simp+


(* Document Definition *)

datatype doc = Doc "path \<rightharpoonup> value"

fun doc_read :: "doc \<Rightarrow> path \<rightharpoonup> value" ("_'(_')\<^sub>1" [1000, 0] 1000) where
  "(Doc d)(p)\<^sub>1 = d p"

theorem doc_read_partial [simp]:
  shows "doc_read (Doc d) = d"
  by (rule ext) simp

abbreviation doc_dom :: "doc \<Rightarrow> path set" where
  "doc_dom d \<equiv> dom (doc_read d)"

abbreviation finiteD :: "doc \<Rightarrow> bool" where
  "finiteD d \<equiv> finite (doc_dom d)"

abbreviation dom_elem :: "path \<Rightarrow> doc \<Rightarrow> bool" ("_ \<in>\<^sub>1 _" [1000, 1000] 100) where
  "p \<in>\<^sub>1 d \<equiv> p \<in> doc_dom d"

abbreviation dom_elem_not :: "path \<Rightarrow> doc \<Rightarrow> bool" ("_ \<notin>\<^sub>1 _" [1000, 1000] 100) where
  "p \<notin>\<^sub>1 d \<equiv> p \<notin> doc_dom d"

definition doc_upd :: "doc \<Rightarrow> path \<Rightarrow> value \<Rightarrow> doc" ("_[_ \<rightarrow> _]" [1000, 0, 0] 1000) where
  "d[p \<rightarrow> v] \<equiv> Doc (doc_read d(p \<mapsto> v))"

theorem read_upd:
  shows "d[p \<rightarrow> v](q)\<^sub>1 = (if p \<noteq> q then d(q)\<^sub>1 else v\<^sub>!)"
  unfolding doc_upd_def by simp

corollary read_upd_simps [simp]:
  shows "d[p \<rightarrow> v](p)\<^sub>1 = v\<^sub>!"
    and "p \<noteq> q \<Longrightarrow> d[p \<rightarrow> v](q)\<^sub>1 = d(q)\<^sub>1"
  unfolding read_upd by simp+

definition doc_del :: "doc \<Rightarrow> path \<Rightarrow> doc" ("_[_ \<rightarrow> ]" [1000, 0] 1000) where
  "d[p \<rightarrow> ] \<equiv> Doc (\<lambda>q. if p \<in> anc q then None else d(q)\<^sub>1)"

theorem read_del:
  shows "d[p \<rightarrow> ](q)\<^sub>1 = (if q \<notin> desc p then d(q)\<^sub>1 else None)"
  unfolding doc_del_def by simp

corollary read_del_simps [simp]:
  shows "q \<in> desc p \<Longrightarrow> d[p \<rightarrow> ](q)\<^sub>1 = None"
    and "q \<notin> desc p \<Longrightarrow> d[p \<rightarrow> ](q)\<^sub>1 = d(q)\<^sub>1"
  unfolding read_del by simp+

definition treeshaped :: "doc \<Rightarrow> bool" where
  "treeshaped d \<equiv> \<forall>p. p \<in>\<^sub>1 d \<longrightarrow> p/.. \<in>\<^sub>1 d"

abbreviation document :: "doc \<Rightarrow> bool" where
  "document d \<equiv> finiteD d \<and> treeshaped d"

lemmas doc_defs = doc_upd_def doc_del_def treeshaped_def


(* Properties *)

declare doc_defs [simp]

abbreviation emptyDoc :: doc where
  "emptyDoc \<equiv> Doc [Root \<mapsto> Unit]"

theorem emptyDoc_document [simp]:
  shows "document emptyDoc"
  by simp

theorem doc_Doc_fromDoc [simp]:
  shows "Doc (doc_read d) = d"
  by (cases d) simp

theorem doc_upd_finite_dom [simp]:
  shows "finiteD d[p \<rightarrow> v] = finiteD d"
  using assms by simp

theorem doc_del_finite_dom [simp]:
  assumes "finiteD d"
  shows "finiteD d[p \<rightarrow>]"
  using assms by - (simp, erule finite_subset[rotated], clarsimp, case_tac "p \<in> anc x", auto)

theorem doc_read_finite_range [simp]:
  assumes "finiteD d"
  shows "finite {p. d(p)\<^sub>1 = x\<^sub>!}"
  using assms by - (erule finite_subset[rotated], force simp add: dom_def)

theorem doc_upd_treeshaped [simp]:
  assumes "p/.. \<in>\<^sub>1 d" and "treeshaped d"
  shows "treeshaped d[p \<rightarrow> v]"
  using assms by simp

theorem doc_del_treeshaped [simp]:
  assumes "treeshaped d"
  shows "treeshaped d[p \<rightarrow>]"
  using assms by (clarsimp, case_tac "p \<in> anc pa", auto)

theorem doc_upd_document [simp]:
  assumes "p/.. \<in>\<^sub>1 d" and "document d"
  shows "finiteD d[p \<rightarrow> v]" and "treeshaped d[p \<rightarrow> v]"
  using assms by simp+

theorem doc_del_document [simp]:
  assumes "document d"
  shows "finiteD d[p \<rightarrow>]" and "treeshaped d[p \<rightarrow>]"
  using assms by (simp del: doc_defs)+

theorem document_parent_dom [simp]:
  assumes "document d" and "p \<in>\<^sub>1 d"
  shows "p/.. \<in>\<^sub>1 d"
  using assms by simp


declare doc_defs [simp del]

end

