header {* Semantic Domains and Lifting Functions *}

theory Domain imports VIPSet begin


(*--------------------*)
(*- Semantic Domains -*)
(*--------------------*)

datatype "domain" =
  DValue  "value mset" ("_\<^sub>v" [1000] 1000)
| DPath   "path mset"  ("_\<^sub>p" [1000] 1000)
| DDoc    doc          ("_\<^sub>d" [1000] 1000)
| DBool   bool         ("_\<^sub>b" [1000] 1000)
| DBot                 ("\<bottom>")       

declare [[coercion DDoc]]
declare [[coercion DBool]]


(*----------------*)
(*- Base Domains -*)
(*----------------*)

class basedomain =
fixes castM :: "domain \<rightharpoonup> 'a mset"
  and consM :: "'a mset \<Rightarrow> domain"
assumes consM_inj [simp]:
  "inj consM"
and castM_mscon [simp]:
  "castM (consM x) = x\<^sub>!"
and castM_inj [simp]:
  "partial_inj castM"
and castM_strict [simp]:
  "castM \<bottom> = None"

theorem basedomain_cast_simp [simp]:
  fixes y :: "('a :: basedomain) mset"
  shows "castM x = y\<^sub>! \<longleftrightarrow> x = consM y"
  by auto (metis castM_mscon castM_inj partial_inj)


instantiation "value" :: basedomain begin

  fun castM_value where
    "castM_value (DValue m) = m\<^sub>!"
  | "castM_value _ = None"

  definition consM_value where
    "consM_value \<equiv> DValue"

instance by (intro_classes, auto simp: consM_value_def partial_inj intro: injI, case_tac x, auto, case_tac y, auto) end

declare consM_value_def [simp]


instantiation "path" :: basedomain begin

  fun castM_path where
    "castM_path (DPath m) = m\<^sub>!"
  | "castM_path _ = None"

  definition consM_path where
    "consM_path \<equiv> DPath"

instance by (intro_classes, auto simp: consM_path_def partial_inj intro: injI, case_tac x, auto, case_tac y, auto) end

declare consM_path_def [simp]


(*-----------*)
(*- Domains -*)
(*-----------*)

class "domain" =
fixes cast :: "domain \<rightharpoonup> 'a"
  and cons :: "'a \<Rightarrow> domain"
assumes cons_inj [simp]:
  "inj cons"
and cast_inj [simp]:
  "partial_inj cast"
and cast_strict [simp]:
  "cast \<bottom> = None"
and cast_cons_weak [simp]:
  "cons x \<noteq> \<bottom> \<Longrightarrow> cast (cons x) = y\<^sub>! \<longleftrightarrow> x = y"

corollary cast_cons_dest [dest]:
  assumes "cast (cons x) = y\<^sub>!"
  shows "x = y"
  using assms by (cases "cons x = \<bottom>") auto


instantiation "domain" :: "domain" begin

  fun cast_domain where
    "cast_domain \<bottom> = None"
  | "cast_domain d = Some d"

  definition cons_domain :: "domain \<Rightarrow> domain" where
    "cons_domain \<equiv> id"

instance proof
  show "inj (cons :: domain \<Rightarrow> domain)" unfolding cons_domain_def by simp next
  show "partial_inj (cast :: domain \<Rightarrow> domain option)"
    unfolding partial_inj by (clarify, case_tac x, auto) (case_tac y, auto)+ next
  show "(cast :: domain \<Rightarrow> domain option) \<bottom> = None" by simp next
  fix x :: "domain" and y :: "domain" assume "cons x \<noteq> \<bottom>"
  thus "cast (cons x) = y\<^sub>! \<longleftrightarrow> x = y" unfolding cons_domain_def by (cases x) auto
qed end

declare cons_domain_def [simp]


(*----------------------*)
(*- Proper Sub Domains -*)
(*----------------------*)

class subdomain = "domain" +
assumes cast_cons [simp]: "cast (cons x) = x\<^sub>!"

theorem subdomain_cons_defined_left [simp]:
  fixes x :: "'a :: subdomain"
  shows "cons x \<noteq> \<bottom>"
proof
  assume "cons x = \<bottom>" hence "cast (cons x) = None" by simp
  thus False by (subst (asm) cast_cons) simp
qed

theorem subdomain_cons_defined_right [simp]:
  fixes x :: "'a :: subdomain"
  shows "\<bottom> \<noteq> cons x"
proof
  assume "\<bottom> = cons x" hence "cast (cons x) = None" by simp
  thus False by (subst (asm) cast_cons) simp
qed

theorem subdomain_cast_simp:
  fixes y :: "'a :: subdomain"
  shows "cast x = y\<^sub>! \<longleftrightarrow> x = cons y"
  by auto (metis cast_cons cast_inj partial_inj)


instantiation mset :: (basedomain) subdomain begin

  definition cast_mset where
    "cast_mset \<equiv> castM"
  definition cons_mset where
    "cons_mset \<equiv> consM"

instance by (intro_classes, auto simp: cast_mset_def cons_mset_def) end

declare cast_mset_def [simp]
declare cons_mset_def [simp]


instantiation doc :: subdomain begin

  fun cast_doc where
    "cast_doc (DDoc m) = m\<^sub>!"
  | "cast_doc _ = None"

  definition cons_doc where
    "cons_doc \<equiv> DDoc"

instance by (intro_classes, auto simp: cons_doc_def partial_inj intro: injI, case_tac x, auto, case_tac y, auto) end

declare cons_doc_def [simp]


instantiation bool :: subdomain begin

  fun cast_bool where
    "cast_bool (DBool m) = m\<^sub>!"
  | "cast_bool _ = None"

  definition cons_bool where
    "cons_bool \<equiv> DBool"

instance by (intro_classes, auto simp: cons_bool_def partial_inj intro: injI, case_tac x, auto, case_tac y, auto) end

declare cons_bool_def [simp]


(* All proper subdomains can be lifted to a domain, keeping type safety *)

instantiation option :: ("subdomain") "domain" begin

  fun cons_option :: "'a option \<Rightarrow> domain" where
    "cons_option x\<^sub>! = cons x"
  | "cons_option None = \<bottom>"

  fun cast_option :: "domain \<Rightarrow> 'a option option" where
    "cast_option \<bottom> = None"
  | "cast_option x = Option.map Some (cast x)"

  lemma cast_option_defined [simp]:
    assumes "x \<noteq> \<bottom>"
    shows "(cast x :: ('a :: subdomain) option option) = Option.map Some (cast x)"
    using assms by (cases x) auto

instance proof
  show "inj (cons :: 'a option \<Rightarrow> domain)" using cons_inj[THEN injD]
    by - (rule injI, case_tac x, auto, case_tac y, auto, case_tac y, auto) next
  have "partial_inj (cast :: domain \<Rightarrow> 'a option)" by simp
  thus "partial_inj (cast :: domain \<Rightarrow> 'a option option)" 
    unfolding partial_inj by clarify (case_tac "x = \<bottom>", auto, case_tac "y = \<bottom>", auto) next
  show "cast \<bottom> = (None :: 'a option option)" by simp next
  fix x :: "'a option" and y :: "'a option" assume "cons x \<noteq> \<bottom>"
  thus "cast (cons x) = y\<^sub>! \<longleftrightarrow> x = y" by (cases x) auto
qed end

declare subdomain_cast_simp [simp]  (* instance proof loops with this rule *)


(*---------------------*)
(*- Sub Domain Values -*)
(*---------------------*)

definition theDomain :: "('a :: domain) itself \<Rightarrow> domain set" where
  "theDomain t \<equiv> insert \<bottom> (dom (cast :: domain \<Rightarrow> 'a option))"

theorem theDomainD [dest!]:
  fixes t :: "('a :: domain) itself"
  assumes "x \<in> theDomain t"
  shows "x = \<bottom> \<or> (\<exists>y :: 'a. cast x = y\<^sub>!)"
  using assms unfolding theDomain_def by auto

corollary theDomain_cast_none [simp]:
  fixes t :: "('a :: domain) itself"
  assumes "x \<in> theDomain TYPE('a)"
  shows "cast x = (None :: 'a option) \<longleftrightarrow> x = \<bottom>"
  using assms by auto

corollary theDomain_castM_none [simp]:
  fixes t :: "('a :: basedomain) mset itself"
  assumes "x \<in> theDomain TYPE('a mset)"
  shows "castM x = (None :: 'a mset option) \<longleftrightarrow> x = \<bottom>"
  using assms by auto


abbreviation valueDom where
  "valueDom \<equiv> theDomain TYPE(value mset)"

abbreviation pathDom where
  "pathDom \<equiv> theDomain TYPE(path mset)"

abbreviation docDom where
  "docDom \<equiv> theDomain TYPE(doc)"

abbreviation boolDom where
  "boolDom \<equiv> theDomain TYPE(bool)"


theorem theDomain_bot [simp]:
  shows "\<bottom> \<in> theDomain t"
  unfolding theDomain_def by simp

theorem theDomain_cons [simp]:
  fixes x :: "'a :: domain" and t :: "'a itself"
  shows "cons x \<in> theDomain t"
  unfolding theDomain_def by auto

corollary theDomain_cons_option [simp]:
  fixes x :: "('a :: subdomain) option" and t :: "'a itself"
  shows "cons x \<in> theDomain t" by (cases x) auto

corollary theDomain_value [simp]:
  shows "x\<^sub>v \<in> valueDom"
  by (metis consM_value_def cons_mset_def theDomain_cons)

corollary theDomain_path [simp]:
  shows "x\<^sub>p \<in> pathDom"
  by (metis consM_path_def cons_mset_def theDomain_cons)

corollary theDomain_doc [simp]:
  shows "x\<^sub>d \<in> docDom"
  by (metis cons_doc_def theDomain_cons)

corollary theDomain_bool [simp]:
  shows "x\<^sub>b \<in> boolDom"
  by (metis cons_bool_def theDomain_cons)


(*--------------*)
(*- Finiteness -*)
(*--------------*)

primrec finiteO :: "domain \<Rightarrow> bool" ("finiteO _" [999] 999) where  (* so we can write finiteO \<parallel>x\<parallel>\<^sub>E later on *)
  "finiteO (DValue v) = finiteM v"
| "finiteO (DPath p)  = finiteM p"
| "finiteO (DDoc d)   = finiteD d"
| "finiteO (DBool b)  = True"
| "finiteO DBot       = True"


(*-------------------*)
(*- Domain Liftings -*)
(*-------------------*)

(* One Parameter *)

definition liftDom :: "('a :: subdomain \<Rightarrow> 'b :: domain) \<Rightarrow> domain \<Rightarrow> domain" where
  "liftDom f a \<equiv> (case cast a of da\<^sub>! \<Rightarrow> cons (f da) | _ \<Rightarrow> \<bottom>)"

theorem liftDom_strict [simp]:
  shows "liftDom f \<bottom> = \<bottom>"
  unfolding liftDom_def by simp

theorem liftDom_cast_strict [simp]: 
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: domain"
  assumes "cast a = (None :: 'a option)"
  shows "liftDom f a = \<bottom>"
  using assms unfolding liftDom_def by simp

theorem liftDom_app [simp]:
  assumes "x = cons a"
  shows "liftDom f x = cons (f a)"
  using assms unfolding liftDom_def by simp

theorem liftDom_undefined [elim!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: domain"
  assumes "liftDom f a = \<bottom>" and "\<And>x. \<lbrakk> cast a = x\<^sub>!; cons (f x) = \<bottom> \<rbrakk> \<Longrightarrow> P"
    and "cast a = (None :: 'a option) \<Longrightarrow> P"
  shows P
  using assms unfolding liftDom_def by (cases "cast a :: 'a option") auto

theorem liftDom_undefinedI [intro!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: domain"
  assumes "\<And>x. cast a = x\<^sub>! \<Longrightarrow> cons (f x) = \<bottom>"
  shows "liftDom f a = \<bottom>"
  using assms unfolding liftDom_def by (cases "cast a :: 'a option") auto

corollary liftDom_defined [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: domain"
  assumes "liftDom f a \<noteq> \<bottom>"
  shows "\<exists>x. a = cons x \<and> cons (f x) \<noteq> \<bottom>"
  using assms by auto

theorem liftDom_defined_cons:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: domain" and r :: "'c :: subdomain"
  assumes "liftDom f a = cons r"
  shows "\<exists>x. a = cons x \<and> cons (f x) = cons r"
proof -
  from assms have "liftDom f a \<noteq> \<bottom>" by simp
  then obtain x :: 'a where "a = cons x"
    using liftDom_defined[of f a] by auto moreover
  hence "cons (f x) = cons r" using assms unfolding liftDom_def by simp ultimately
  show ?thesis by auto
qed

corollary liftDom_defined_value [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: domain"
  assumes "liftDom f a = r\<^sub>v"
  shows "\<exists>x. a = cons x \<and> cons (f x) = r\<^sub>v"
  using assms and liftDom_defined_cons[of f a r] by simp

corollary liftDom_defined_path [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: domain"
  assumes "liftDom f a = r\<^sub>p"
  shows "\<exists>x. a = cons x \<and> cons (f x) = r\<^sub>p"
  using assms and liftDom_defined_cons[of f a r] by simp

corollary liftDom_defined_doc [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: domain"
  assumes "liftDom f a = r\<^sub>d"
  shows "\<exists>x. a = cons x \<and> cons (f x) = r\<^sub>d"
  using assms and liftDom_defined_cons[of f a r] by simp

corollary liftDom_defined_bool [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: domain"
  assumes "liftDom f a = r\<^sub>b"
  shows "\<exists>x. a = cons x \<and> cons (f x) = r\<^sub>b"
  using assms and liftDom_defined_cons[of f a r] by simp


(* Two Parameters *)

definition liftDom2 ::
  "('a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: domain) \<Rightarrow>
             domain \<Rightarrow>          domain \<Rightarrow>       domain" where
  "liftDom2 f a b = (case (cast a, cast b) of (da\<^sub>!, db\<^sub>!) \<Rightarrow> cons (f da db) | _ \<Rightarrow> \<bottom>)"

theorem liftDom2_strict_left [simp]:
  shows "liftDom2 f \<bottom> b = \<bottom>"
  unfolding liftDom2_def by simp

theorem liftDom2_strict_right [simp]:
  shows "liftDom2 f a \<bottom> = \<bottom>"
  unfolding liftDom2_def by simp (cases "cast a :: 'a option", auto)

theorem liftDom2_cast_strict_left [simp]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: domain"
  assumes "cast a = (None :: 'a option)"
  shows "liftDom2 f a b = \<bottom>"
  using assms unfolding liftDom2_def by simp 

theorem liftDom2_cast_strict_right [simp]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: domain"
  assumes "cast b = (None :: 'b option)"
  shows "liftDom2 f a b = \<bottom>"
  using assms unfolding liftDom2_def by simp (cases "cast a :: 'a option", auto)

theorem liftDom2_app [simp]:
  assumes "x = cons a" and "y = cons b"
  shows "liftDom2 f x y = cons (f a b)"
  using assms unfolding liftDom2_def by simp

theorem liftDom2_undefined [elim!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: domain"
  assumes "liftDom2 f a b = \<bottom>" and "\<And>x y. \<lbrakk> cast a = x\<^sub>!; cast b = y\<^sub>!; cons (f x y) = \<bottom> \<rbrakk> \<Longrightarrow> P"
    and "cast a = (None :: 'a option) \<Longrightarrow> P" and "cast b = (None :: 'b option) \<Longrightarrow> P"
  shows P
  using assms unfolding liftDom2_def
  by (cases "cast a :: 'a option", auto, cases "cast b :: 'b option") auto

theorem liftDom2_undefined_intro [intro!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: domain"
  assumes "\<And>x y. \<lbrakk> cast a = x\<^sub>!; cast b = y\<^sub>! \<rbrakk> \<Longrightarrow> cons (f x y) = \<bottom>"
  shows "liftDom2 f a b = \<bottom>"
  using assms unfolding liftDom2_def
  by (cases "cast a :: 'a option", auto, cases "cast b :: 'b option") auto

corollary liftDom2_defined [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: domain"
  assumes "liftDom2 f a b \<noteq> \<bottom>"
  shows "\<exists>x y. a = cons x \<and> b = cons y \<and> cons (f x y) \<noteq> \<bottom>"
  using assms by auto

theorem liftDom2_defined_cons:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: domain" and r :: "'d :: subdomain"
  assumes "liftDom2 f a b = cons r"
  shows "\<exists>x y. a = cons x \<and> b = cons y \<and> cons (f x y) = cons r"
proof -
  from assms have "liftDom2 f a b \<noteq> \<bottom>" by simp
  then obtain x :: 'a and y :: 'b where "a = cons x" and "b = cons y"
    using liftDom2_defined[of f a b] by auto moreover
  hence "cons (f x y) = cons r" using assms unfolding liftDom2_def by simp ultimately
  show ?thesis by auto
qed

corollary liftDom2_defined_path [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: domain"
  assumes "liftDom2 f a b = r\<^sub>p"
  shows "\<exists>x y. a = cons x \<and> b = cons y \<and> cons (f x y) = r\<^sub>p"
  using assms and liftDom2_defined_cons[of f a b r] by simp

corollary liftDom2_defined_value [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: domain"
  assumes "liftDom2 f a b = r\<^sub>v"
  shows "\<exists>x y. a = cons x \<and> b = cons y \<and> cons (f x y) = r\<^sub>v"
  using assms and liftDom2_defined_cons[of f a b r] by simp

corollary liftDom2_defined_doc [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: domain"
  assumes "liftDom2 f a b = r\<^sub>d"
  shows "\<exists>x y. a = cons x \<and> b = cons y \<and> cons (f x y) = r\<^sub>d"
  using assms and liftDom2_defined_cons[of f a b r] by simp

corollary liftDom2_defined_bool [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: domain"
  assumes "liftDom2 f a b = r\<^sub>b"
  shows "\<exists>x y. a = cons x \<and> b = cons y \<and> cons (f x y) = r\<^sub>b"
  using assms and liftDom2_defined_cons[of f a b r] by simp


(* Three Parameters *)

definition liftDom3 ::
  "('a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: subdomain \<Rightarrow> 'd :: domain) \<Rightarrow>
             domain \<Rightarrow>          domain \<Rightarrow>          domain \<Rightarrow>       domain" where
  "liftDom3 f a b c = (case (cast a, cast b, cast c) of
    (da\<^sub>!, db\<^sub>!, dc\<^sub>!) \<Rightarrow> cons (f da db dc) | _ \<Rightarrow> \<bottom>)"

theorem liftDom3_strict_left [simp]:
  shows "liftDom3 f \<bottom> b c = \<bottom>"
  unfolding liftDom3_def by simp

theorem liftDom3_strict_middle [simp]:
  shows "liftDom3 f a \<bottom> c = \<bottom>"
  unfolding liftDom3_def by simp (cases "cast a :: 'a option", auto)

theorem liftDom3_strict_right [simp]:
  shows "liftDom3 f a b \<bottom> = \<bottom>"
  unfolding liftDom3_def
  by simp (cases "cast a :: 'a option", auto, cases "cast b :: 'b option", auto)

theorem liftDom3_correct_application [simp]:
  assumes "x = cons a" and "y = cons b" and "z = cons c"
  shows "liftDom3 f x y z = cons (f a b c)"
  using assms unfolding liftDom3_def by simp

theorem liftDom3_cast_strict_left [simp]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: subdomain \<Rightarrow> 'd :: domain"
  assumes "cast a = (None :: 'a option)"
  shows "liftDom3 f a b c = \<bottom>"
  using assms unfolding liftDom3_def by simp

theorem liftDom3_cast_strict_middle [simp]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: subdomain \<Rightarrow> 'd :: domain"
  assumes "cast b = (None :: 'b option)"
  shows "liftDom3 f a b c = \<bottom>"
  using assms unfolding liftDom3_def by simp (cases "cast a :: 'a option", auto)

theorem liftDom3_cast_strict_right [simp]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: subdomain \<Rightarrow> 'd :: domain"
  assumes "cast c = (None :: 'c option)"
  shows "liftDom3 f a b c = \<bottom>"
  using assms unfolding liftDom3_def
  by simp (cases "cast a :: 'a option", auto, cases "cast b :: 'b option", auto)

theorem liftDom3_undefined [elim!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: subdomain \<Rightarrow> 'd :: domain"
  assumes "liftDom3 f a b c = \<bottom>"
    and  "\<And>x y z. \<lbrakk> cast a = x\<^sub>!; cast b = y\<^sub>!; cast c = z\<^sub>!; cons (f x y z) = \<bottom> \<rbrakk> \<Longrightarrow> P"
    and "cast a = (None :: 'a option) \<Longrightarrow> P"
    and "cast b = (None :: 'b option) \<Longrightarrow> P"
    and "cast c = (None :: 'c option) \<Longrightarrow> P"
  shows P
  using assms unfolding liftDom2_def
  by (cases "cast a :: 'a option", auto, cases "cast b :: 'b option", auto)
     (cases "cast c :: 'c option", auto)

theorem liftDom3_undefined_intro [intro!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: subdomain \<Rightarrow> 'd :: domain"
  assumes "\<And>x y z. \<lbrakk> cast a = x\<^sub>!; cast b = y\<^sub>!; cast c = z\<^sub>! \<rbrakk> \<Longrightarrow> cons (f x y z) = \<bottom>"
  shows "liftDom3 f a b c = \<bottom>"
  using assms unfolding liftDom3_def
  by (cases "cast a :: 'a option", auto, cases "cast b :: 'b option", auto)
     (cases "cast c :: 'c option", auto)

corollary liftDom3_defined [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: subdomain \<Rightarrow> 'd :: domain"
  assumes "liftDom3 f a b c \<noteq> \<bottom>"
  shows "\<exists>x y z. a = cons x \<and> b = cons y \<and> c = cons z \<and> cons (f x y z) \<noteq> \<bottom>"
  using assms by auto

theorem liftDom3_defined_cons:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: subdomain \<Rightarrow> 'd :: domain"
    and r :: "'e :: subdomain"
  assumes "liftDom3 f a b c = cons r"
  shows "\<exists>x y z. a = cons x \<and> b = cons y \<and> c = cons z \<and> cons (f x y z) = cons r"
proof -
  from assms have "liftDom3 f a b c \<noteq> \<bottom>" by simp
  then obtain x :: 'a and y :: 'b and z :: 'c where "a = cons x" and "b = cons y" and "c = cons z"
    using liftDom3_defined[of f a b c] by auto moreover
  hence "cons (f x y z) = cons r" using assms unfolding liftDom3_def by simp ultimately
  show ?thesis by auto
qed

corollary liftDom3_defined_path [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: subdomain \<Rightarrow> 'd :: domain"
  assumes "liftDom3 f a b c = r\<^sub>p"
  shows "\<exists>x y z. a = cons x \<and> b = cons y \<and> c = cons z \<and> cons (f x y z) = r\<^sub>p"
  using assms and liftDom3_defined_cons[of f a b c r] by simp

corollary liftDom3_defined_value [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: subdomain \<Rightarrow> 'd :: domain"
  assumes "liftDom3 f a b c = r\<^sub>v"
  shows "\<exists>x y z. a = cons x \<and> b = cons y \<and> c = cons z \<and> cons (f x y z) = r\<^sub>v"
  using assms and liftDom3_defined_cons[of f a b c r] by simp

corollary liftDom3_defined_doc [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: subdomain \<Rightarrow> 'd :: domain"
  assumes "liftDom3 f a b c = r\<^sub>d"
  shows "\<exists>x y z. a = cons x \<and> b = cons y \<and> c = cons z \<and> cons (f x y z) = r\<^sub>d"
  using assms and liftDom3_defined_cons[of f a b c r] by simp

corollary liftDom3_defined_bool [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: subdomain \<Rightarrow> 'd :: domain"
  assumes "liftDom3 f a b c = r\<^sub>b"
  shows "\<exists>x y z. a = cons x \<and> b = cons y \<and> c = cons z \<and> cons (f x y z) = r\<^sub>b"
  using assms and liftDom3_defined_cons[of f a b c r] by simp


(* theDomain Theorems *)

theorem theDomain_liftDom [simp]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: domain" and t :: "'b itself"
  shows "liftDom f a \<in> theDomain t"
  unfolding liftDom_def by (cases "cast a :: 'a option") auto

theorem theDomain_liftDom2 [simp]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: domain" and t :: "'c itself"
  shows "liftDom2 f a b \<in> theDomain t"
  unfolding liftDom2_def by (cases "cast a :: 'a option", auto, cases "cast b :: 'b option", auto)

theorem theDomain_liftDom3 [simp]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: subdomain \<Rightarrow> 'd :: domain" 
    and t :: "'d itself"
  shows "liftDom3 f a b c \<in> theDomain t"
  unfolding liftDom3_def
  by (cases "cast a :: 'a option", auto, cases "cast b :: 'b option", auto)
     (cases "cast c :: 'c option", auto)


theorem theDomain_liftDom_option [simp]:
  fixes f :: "'a :: subdomain \<Rightarrow> ('b :: subdomain) option" and t :: "'b itself"
  shows "liftDom f a \<in> theDomain t"
  unfolding liftDom_def by (cases "cast a :: 'a option") auto

theorem theDomain_liftDom2_option [simp]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> ('c :: subdomain) option" and t :: "'c itself"
  shows "liftDom2 f a b \<in> theDomain t"
  unfolding liftDom2_def by (cases "cast a :: 'a option", auto, cases "cast b :: 'b option", auto)

theorem theDomain_liftDom3_option [simp]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: subdomain \<Rightarrow> ('d :: subdomain) option" 
    and t :: "'d itself"
  shows "liftDom3 f a b c \<in> theDomain t"
  unfolding liftDom3_def
  by (cases "cast a :: 'a option", auto, cases "cast b :: 'b option", auto)
     (cases "cast c :: 'c option", auto)


(* Better Cast Theorems *)

theorem cast_liftDom_defined [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: domain" and r :: 'b
  assumes "cast (liftDom f a) = r\<^sub>!"
  shows "liftDom f a = cons r"
  using assms by (cases "cast a :: 'a option", auto)

corollary cast_liftDom2_defined [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: domain" and r :: 'c
  assumes "cast (liftDom2 f a b) = r\<^sub>!"
  shows "liftDom2 f a b = cons r"
  using assms by (cases "cast a :: 'a option", auto, cases "cast b :: 'b option", auto)

corollary cast_liftDom3_defined [dest!]:
  fixes f :: "'a :: subdomain \<Rightarrow> 'b :: subdomain \<Rightarrow> 'c :: subdomain \<Rightarrow> 'd :: domain" and r :: 'd
  assumes "cast (liftDom3 f a b c) = r\<^sub>!"
  shows "liftDom3 f a b c = cons r"
  using assms
  by (cases "cast a :: 'a option", auto, cases "cast b :: 'b option", auto)
     (cases "cast c :: 'c option", auto)


end
