header {* Syntax and Semantics of Expressions *}

theory Expressions imports Domain begin


(* Syntax *)

datatype name = Name string

datatype valueExp =
  Val "value"
| FVar name
| PVar name

| Sum    valueExp
| Size   valueExp
| Count  valueExp valueExp
| Tally  type     valueExp

| Read   docExp   pathExp

| Scal   valueExp valueExp
| VRep   valueExp valueExp
| VPlus  valueExp valueExp
| VMinus valueExp valueExp

| VCond lit       valueExp valueExp

and pathExp =
  PRoot
| Path   label    pathExp valueExp
| Kind   label    pathExp

| PRep   valueExp pathExp
| PPlus  pathExp  pathExp
| PMinus pathExp  pathExp

| PCond lit       pathExp pathExp

and docExp =
  Init
| Upd docExp pathExp valueExp
| Del docExp pathExp

| DCond lit  docExp  docExp

and pred = 
  Equal valueExp valueExp
| Less  valueExp valueExp

| RCond lit pred pred

and lit = L bool pred


(* Important Constants *)

abbreviation exp_0 :: "valueExp" ("0.") where
  "0. \<equiv> Val (VInt 0)"
abbreviation exp_1 :: "valueExp" ("1.") where
  "1. \<equiv> Val (VInt 1)"
abbreviation exp_1_neg :: "valueExp" ("~1.") where
  "~1. \<equiv> Val (VInt (-1))"

abbreviation exp_null :: "valueExp" ("null.") where
  "null. \<equiv> Val null"

abbreviation exp_empty :: "valueExp" ("{}.") where
  "{}. \<equiv> Scal 0. null."
abbreviation exp_bot :: "valueExp" ("\<bottom>.") where
  "\<bottom>. \<equiv> Scal null. 0."

abbreviation path_root :: "pathExp" ("Root.") where
  "Root. \<equiv> PRoot"

abbreviation doc_init :: "docExp" ("$.") where
  "$. \<equiv> Init"


class boolExp =
fixes bFalse ("False.")
  and bTrue  ("True.")

instantiation pred and lit :: boolExp begin

  definition bTrue_pred  :: pred where
    "True.  \<equiv> Equal 0. 0."
  definition bTrue_lit   :: lit  where
    "True.  \<equiv> L True True. "
  definition bFalse_pred :: pred where
    "False. \<equiv> Less 0. 0."
  definition bFalse_lit :: lit  where
    "False. \<equiv> L True False."

instance by intro_classes end

lemmas bexp_defs = bTrue_pred_def bTrue_lit_def bFalse_pred_def bFalse_lit_def 
lemmas bexp_normalize [symmetric, simp] = bexp_defs


(* Kind Instances *)

instantiation pathExp :: kindOf begin

  fun kindOf_pathExp :: "pathExp \<rightharpoonup> kind" where
    "kindOf Root.        = EmptyKind\<^sub>!"
  | "kindOf (Path l p k) = Option.map (Step l) (kindOf p)"
  | "kindOf (Kind l p)   = Option.map (Step l) (kindOf p)"

  | "kindOf (PRep n p)   = kindOf p"
  | "kindOf (PPlus p q)  = kindOf p"
  | "kindOf (PMinus p q) = kindOf p"

  | "kindOf (PCond b t e) = kindOf t"

instance by intro_classes end

theorem kindOf_pathExp_not_none [simp]:
  fixes p :: pathExp
  shows "kindOf p \<noteq> None"
  by (induct p) auto

instantiation valueExp and docExp and pred and lit :: kindOf begin

  fun kindOf_valueExp and kindOf_docExp and kindOf_pred and kindOf_lit where
    "kindOf (_ :: valueExp) = None" 
  | "kindOf (_ :: docExp)   = None" 
  | "kindOf (_ :: pred)     = None" 
  | "kindOf (_ :: lit)      = None" 

instance by intro_classes end


(* Specialized Semantic Functions *)

definition "assume" :: "bool \<Rightarrow> 'a \<rightharpoonup> 'a" ("[assume: _] _" [0, 100] 100) where
  "[assume: b] x \<equiv> if b then x\<^sub>! else None"

theorem assume_some [simp]:
  shows "[assume: b] x = y\<^sub>! \<longleftrightarrow> b \<and> y = x"
  unfolding assume_def by auto

theorem assume_none [simp]:
  shows "[assume: b] x = None \<longleftrightarrow> \<not>b"
  unfolding assume_def by auto

theorem assume_true_false [simp]:
  shows "[assume: True] x = x\<^sub>!"
    and "[assume: False] x = None"
  by simp+

corollary assume_undefined [simp]:
  shows "cons ([assume: b] x) = \<bottom> \<longleftrightarrow> \<not> b "
  by auto

corollary assume_defined [simp]:
  assumes "cons y = z"
  shows "cons ([assume: b] x) = z \<longleftrightarrow> b \<and> x = y"
  using assms by (cases b) (auto elim: injD[rotated])


lemma int_eq_1 [simp]:
  shows "int x = 1 \<longleftrightarrow> x = 1"
  by auto


abbreviation semSum :: "value mset \<Rightarrow> value mset" where
  "semSum m \<equiv> {sumM (asIntM m)}\<^sub>i"

abbreviation semSize :: "'a mset \<Rightarrow> value mset" where
  "semSize m \<equiv> {int (sizeM m)}\<^sub>i"

abbreviation semTally :: "type \<Rightarrow> value mset \<Rightarrow> value mset" where
  "semTally t m \<equiv> semSize (m\<down>(valuesOf t))"


abbreviation semSumSafe :: "value mset \<Rightarrow> value mset option" where
  "semSumSafe m \<equiv> [assume: finiteM m] semSum m"

abbreviation semSizeSafe :: "'a mset \<Rightarrow> value mset option" where
  "semSizeSafe m \<equiv> [assume: finiteM m] semSize m"

abbreviation semTallySafe :: "type \<Rightarrow> value mset \<Rightarrow> value mset option" where
  "semTallySafe t m \<equiv> [assume: finiteM m] semTally t m"


abbreviation semCount :: "'a mset \<Rightarrow> 'a mset \<Rightarrow> value mset option" where
  "semCount m e \<equiv> [assume: unique e] {int (count m e\<^sub>1)}\<^sub>i"


abbreviation semScalar :: "value mset \<Rightarrow> value mset \<Rightarrow> value mset option " where
  "semScalar a b \<equiv> [assume: unique a \<and> a\<^sub>1 \<in> ints] VIntM (theInt a\<^sub>1 \<odot> asIntM b)"

abbreviation semRep :: "value mset \<Rightarrow> ('a :: basedomain) mset \<Rightarrow> 'a mset option" where
  "semRep n m \<equiv> [assume: unique n \<and> n\<^sub>1 \<in> ints \<and> 0 \<le> theInt n\<^sub>1] nat (theInt n\<^sub>1) \<otimes> m"

abbreviation semMinus :: "('a :: basedomain) mset \<Rightarrow> 'a mset \<Rightarrow> 'a mset option" where
  "semMinus a b \<equiv> [assume: b \<subseteq>\<^sub>* a] a \<ominus> b"


abbreviation semChildren :: "label \<Rightarrow> path mset \<Rightarrow> value mset \<Rightarrow> path mset" where
  "semChildren l p k \<equiv> children p l (asKeyM k)"


abbreviation semUpd :: "doc \<Rightarrow> path mset \<Rightarrow> value mset \<Rightarrow> doc option"  where
  "semUpd d p v \<equiv> [assume: unique p \<and> unique v] d[p\<^sub>1 \<rightarrow> v\<^sub>1]"

abbreviation semDel :: "doc \<Rightarrow> path mset \<Rightarrow> doc option" where
  "semDel d p \<equiv> [assume: unique p] d[p\<^sub>1 \<rightarrow>]"

abbreviation semEqual :: "value mset \<Rightarrow> value mset \<Rightarrow> bool option" where
  "semEqual a b \<equiv> [assume: unique a \<and> unique b] (a = b)"

abbreviation semLess :: "value mset \<Rightarrow> value mset \<Rightarrow> bool option" where
  "semLess a b \<equiv> [assume: unique a \<and> a\<^sub>1 \<in> ints \<and> unique b \<and> b\<^sub>1 \<in> ints] (theInt a\<^sub>1 < theInt b\<^sub>1)"


(* Domain Liftings *)

abbreviation domSumSafe ("sum\<^sub>?\<^sub>? _" [999] 999) where
  "domSumSafe \<equiv> liftDom semSumSafe"

abbreviation domSum ("sum\<^sub>? _" [999] 999) where
  "domSum \<equiv> liftDom semSum"

abbreviation domSizeSafe ("size\<^sub>?\<^sub>? _" [999] 999) where
  "domSizeSafe \<equiv> liftDom (semSizeSafe :: value mset \<Rightarrow> value mset option)"

abbreviation domSize ("size\<^sub>? _" [999] 999) where
  "domSize \<equiv> liftDom (semSize :: value mset \<Rightarrow> value mset)"

abbreviation domTallySafe ("tally\<^sub>?\<^sub>? _ _" [999, 999] 999) where
  "domTallySafe t \<equiv> liftDom (semTallySafe t)"

abbreviation domTally ("tally\<^sub>? _ _" [999, 999] 999) where
  "domTally t \<equiv> liftDom (semTally t)"

abbreviation domCount ("count\<^sub>? _ _" [999, 999] 999) where
  "domCount \<equiv> liftDom2 (semCount :: value mset \<Rightarrow> value mset \<Rightarrow> value mset option)"


abbreviation domMSRead ("_'(_')\<^sub>*\<^sub>?" [999, 0] 999) where
  "domMSRead \<equiv> liftDom2 readM"


abbreviation domScalar ("_ \<odot>\<^sub>? _" [999, 999] 999) where
  "domScalar \<equiv> liftDom2 semScalar"

abbreviation domRepValue ("_ \<otimes>\<^sub>? _" [999, 999] 999) where
  "domRepValue \<equiv> liftDom2 (semRep :: value mset \<Rightarrow> value mset \<Rightarrow> value mset option)"

abbreviation domPlusValue ("_ \<oplus>\<^sub>? _" [999, 999] 999) where
  "domPlusValue \<equiv> liftDom2 (op + :: value mset \<Rightarrow> value mset \<Rightarrow> value mset)"

abbreviation domMinusValue ("_ \<ominus>\<^sub>? _" [999, 999] 999) where
  "domMinusValue \<equiv> liftDom2 (semMinus :: value mset \<Rightarrow> value mset \<Rightarrow> value mset option)"


abbreviation domChildren ("_'/_[_]\<^sub>*\<^sub>?" [999, 0, 0] 999) where
  "domChildren p l v \<equiv> liftDom2 (semChildren l) p v"

abbreviation domRepPath ("_ \<^sub>?\<otimes> _" [999, 999] 999) where
  "domRepPath \<equiv> liftDom2 (semRep :: value mset \<Rightarrow> path mset \<Rightarrow> path mset option)"

abbreviation domPlusPath ("_ \<^sub>?\<oplus> _" [999, 999] 999) where
  "domPlusPath \<equiv> liftDom2 (op + :: path mset \<Rightarrow> path mset \<Rightarrow> path mset)"

abbreviation domMinus ("_ \<^sub>?\<ominus> _" [999, 999] 999) where
  "domMinus \<equiv> liftDom2 (semMinus :: path mset \<Rightarrow> path mset \<Rightarrow> path mset option)"


abbreviation domUpd ("_[_ \<rightarrow> _]\<^sub>?" [999, 0, 0] 999) where
  "domUpd \<equiv> liftDom3 semUpd"

abbreviation domDel ("_[_ \<rightarrow>]\<^sub>?" [999, 0] 999) where
  "domDel \<equiv> liftDom2 semDel"

abbreviation domEqual ("_ =\<^sub>? _" [999, 999] 999) where
  "domEqual \<equiv> liftDom2 semEqual"

abbreviation domLess ("_ <\<^sub>? _" [999, 999] 999) where
  "domLess \<equiv> liftDom2 semLess"


abbreviation domNeg ("\<not>\<^sub>? _" [999] 999) where
  "domNeg \<equiv> liftDom Not"

abbreviation domIff ("_ \<longleftrightarrow>\<^sub>? _" [999, 999] 999) where
  "domIff \<equiv> liftDom2 (op = :: bool \<Rightarrow> bool \<Rightarrow> bool)"

corollary domIff_app [simp]:  (* corollaries for liftDomX_app theorems *)
  shows "a \<longleftrightarrow>\<^sub>? b = (a \<longleftrightarrow> b)"
  by simp

corollary domNeg_app [simp]:  (* the simplifier chokes on False without these corollaries... *)
  shows "\<not>\<^sub>? a = (\<not> a)"
  by simp


(* Conditionals *)

fun domCond :: "domain \<Rightarrow> domain \<Rightarrow> domain \<Rightarrow> domain" ("if\<^sub>? _ then _ else _ fi" [0, 0, 0] 1000) where
  "domCond True  t e = t" (* non-strict! *)
| "domCond False t e = e" (* non-strict! *)
| "domCond _     _ _ = \<bottom>"

theorem domCond_defined [dest!]:
  assumes "if\<^sub>? b then t else e fi \<noteq> \<bottom>"
  shows "b = True \<and> t \<noteq> \<bottom> \<or> b = False \<and> e \<noteq> \<bottom>"
  using assms by (cases "(b, t, e)" rule: domCond.cases) auto

theorem domCond_strict_homomorphism [simp]:
  assumes "g \<bottom> = \<bottom>"
  shows "g (if\<^sub>? b then t else e fi) = if\<^sub>? b then g t else g e fi"
  using assms by (cases b, auto) (case_tac bool, auto)


(* Environments *)

datatype parameter =
  Param doc "name \<Rightarrow> value"

typedef param = "{Param d e | d e. document d}"
  by auto (rule_tac x="Param emptyDoc (VKey \<circ> c\<^sub>N)" in exI, auto)

datatype env = 
  Env param "name \<Rightarrow> key" ("_\<^sub>+\<^sub>_" [1000, 1000] 1000)
| Par param ("_\<^sub>\<up>" [1000] 1000)

declare [[coercion Par]]

fun par :: "env \<Rightarrow> param" ("_\<^sub>R" [1000] 1000) where
  "(R\<^sub>+\<^sub>f)\<^sub>R = R"
| "R\<^sub>R = R"


class assign =
fixes read_doc :: "'a \<Rightarrow> doc" ("_'($')" [1000] 1000)
 and  read_con :: "'a \<Rightarrow> name \<Rightarrow> value" ("_'(_')\<^sub>c" [1000, 0] 1000)
 and  upd_doc :: "'a \<Rightarrow> doc \<Rightarrow> 'a" ("_'(/$ \<mapsto> _')" [1000, 0] 1000)
 and  upd_con :: "'a \<Rightarrow> name \<Rightarrow> value \<Rightarrow> 'a" ("_'(_ \<mapsto>\<^sub>c _')" [1000, 0, 0] 1000)
assumes read_finiteD [simp]:
  "document E($)"
and upd_doc_read [simp]:
  "document d \<Longrightarrow> E($ \<mapsto> d)($) = d"
and upd_doc_read_con [simp]:
  "document d \<Longrightarrow> E($ \<mapsto> d)(n)\<^sub>c = E(n)\<^sub>c"
and upd_con_read [simp]:
  "E(n \<mapsto>\<^sub>c c)(n)\<^sub>c = c"
and upd_con_read_other [simp]:
  "n \<noteq> m \<Longrightarrow> E(n \<mapsto>\<^sub>c c)(m)\<^sub>c = E(m)\<^sub>c"
and upd_con_read_doc [simp]:
  "E(n \<mapsto>\<^sub>c c)($) = E($)"


instantiation param :: assign begin

  definition read_doc_param where
     "E($) \<equiv> case Rep_param E of Param d _ \<Rightarrow> d"
  definition read_con_param where
     "E(n)\<^sub>c \<equiv> case Rep_param E of Param _ e \<Rightarrow> e n"
  definition upd_doc_param where
     "E($ \<mapsto> d) \<equiv> (case Rep_param E of Param _ e \<Rightarrow> Abs_param (Param d e))"
  definition upd_con_param where
     "E(n \<mapsto>\<^sub>c v) \<equiv> (case Rep_param E of Param d e \<Rightarrow> Abs_param (Param d (e(n := v))))"

  lemmas assign_param_defs = read_doc_param_def read_con_param_def upd_doc_param_def upd_con_param_def

instance by intro_classes (case_tac E, auto simp add: Abs_param_inverse assign_param_defs)+ end


instantiation env :: assign begin

  fun read_doc_env where
    "(R\<^sub>+\<^sub>f)($) = R($)"
  | "R\<^sub>\<up>($) = R($)"

  fun read_con_env where
    "(R\<^sub>+\<^sub>f)(n)\<^sub>c = R(n)\<^sub>c"
  | "R\<^sub>\<up>(n)\<^sub>c = R(n)\<^sub>c"

  fun upd_doc_env where
    "(R\<^sub>+\<^sub>f)($ \<mapsto> d) = R($ \<mapsto> d)\<^sub>+\<^sub>f"
  | "R\<^sub>\<up>($ \<mapsto> d) = R($ \<mapsto> d)\<^sub>\<up> "

  fun upd_con_env where
    "(R\<^sub>+\<^sub>f)(n \<mapsto>\<^sub>c v) = R(n \<mapsto>\<^sub>c v)\<^sub>+\<^sub>f"
  | "R\<^sub>\<up>(n \<mapsto>\<^sub>c v) = R(n \<mapsto>\<^sub>c v)\<^sub>\<up>"

instance by intro_classes (case_tac E, auto)+ end

fun read_var :: "env \<Rightarrow> name \<Rightarrow> domain" ("_'(_')\<^sub>w" [1000, 0] 1000) where
  "(P\<^sub>+\<^sub>f)(n)\<^sub>w = {f n}\<^sub>*\<^sub>v"
| "P\<^sub>\<up>(n)\<^sub>w = \<bottom>"

theorem env_upd_doc_read_var [simp]:
  assumes "finiteD d"
  shows "E($ \<mapsto> d)(n)\<^sub>w = E(n)\<^sub>w"
  by (cases E, auto)

theorem env_upd_con_read_var [simp]:
  shows "E(n \<mapsto>\<^sub>c c)(m)\<^sub>w = E(m)\<^sub>w"
  by (cases E, auto)

theorem env_read_var_defined [dest!]:
  assumes "E(n)\<^sub>w \<noteq> \<bottom>"
  shows "\<exists>x. E(n)\<^sub>w = {x}\<^sub>*\<^sub>v"
  using assms by (cases E, auto)


(* Main Semantics Definition *)

class sem =
fixes sem :: "'a \<Rightarrow> env \<Rightarrow> domain" ("\<parallel>_\<parallel>\<^sub>_" [0, 1000] 999)  (* avoid conflicts with R+f envs *)


instantiation valueExp and pathExp and docExp and pred and lit :: sem begin

fun sem_valueExp :: "valueExp \<Rightarrow> env \<Rightarrow> domain"
and sem_pathExp  :: "pathExp  \<Rightarrow> env \<Rightarrow> domain"
and sem_docExp   :: "docExp   \<Rightarrow> env \<Rightarrow> domain"
and sem_pred     :: "pred     \<Rightarrow> env \<Rightarrow> domain"
and sem_lit      :: "lit      \<Rightarrow> env \<Rightarrow> domain" where

  "\<parallel>Val v\<parallel>\<^sub>E       = {v}\<^sub>*\<^sub>v"
| "\<parallel>FVar n\<parallel>\<^sub>E      = E(n)\<^sub>w"
| "\<parallel>PVar n\<parallel>\<^sub>E      = {E(n)\<^sub>c}\<^sub>*\<^sub>v"

| "\<parallel>Sum m\<parallel>\<^sub>E       = sum\<^sub>?\<^sub>? \<parallel>m\<parallel>\<^sub>E"
| "\<parallel>Size m\<parallel>\<^sub>E      = size\<^sub>?\<^sub>? \<parallel>m\<parallel>\<^sub>E"
| "\<parallel>Count m v\<parallel>\<^sub>E   = count\<^sub>? \<parallel>m\<parallel>\<^sub>E \<parallel>v\<parallel>\<^sub>E"
| "\<parallel>Tally t m\<parallel>\<^sub>E   = tally\<^sub>?\<^sub>? t \<parallel>m\<parallel>\<^sub>E"

| "\<parallel>Read d p\<parallel>\<^sub>E    = \<parallel>d\<parallel>\<^sub>E(\<parallel>p\<parallel>\<^sub>E)\<^sub>*\<^sub>?"

| "\<parallel>Scal n m\<parallel>\<^sub>E    = \<parallel>n\<parallel>\<^sub>E \<odot>\<^sub>? \<parallel>m\<parallel>\<^sub>E"
| "\<parallel>VRep n m\<parallel>\<^sub>E    = \<parallel>n\<parallel>\<^sub>E \<otimes>\<^sub>? \<parallel>m\<parallel>\<^sub>E"
| "\<parallel>VPlus a b\<parallel>\<^sub>E   = \<parallel>a\<parallel>\<^sub>E \<oplus>\<^sub>? \<parallel>b\<parallel>\<^sub>E"
| "\<parallel>VMinus a b\<parallel>\<^sub>E  = \<parallel>a\<parallel>\<^sub>E \<ominus>\<^sub>? \<parallel>b\<parallel>\<^sub>E"

| "\<parallel>Root.\<parallel>\<^sub>E       = {Root}\<^sub>*\<^sub>p"
| "\<parallel>Path l p k\<parallel>\<^sub>E  = \<parallel>p\<parallel>\<^sub>E/l[\<parallel>k\<parallel>\<^sub>E]\<^sub>*\<^sub>?"
| "\<parallel>Kind l p\<parallel>\<^sub>E    = \<parallel>p\<parallel>\<^sub>E/l[(\<infinity>\<down>keys)\<^sub>v]\<^sub>*\<^sub>?"

| "\<parallel>PRep n m\<parallel>\<^sub>E    = \<parallel>n\<parallel>\<^sub>E \<^sub>?\<otimes> \<parallel>m\<parallel>\<^sub>E"
| "\<parallel>PPlus a b\<parallel>\<^sub>E   = (if kindOf a = kindOf b then \<parallel>a\<parallel>\<^sub>E \<^sub>?\<oplus> \<parallel>b\<parallel>\<^sub>E else \<bottom>)"
| "\<parallel>PMinus a b\<parallel>\<^sub>E  = \<parallel>a\<parallel>\<^sub>E \<^sub>?\<ominus> \<parallel>b\<parallel>\<^sub>E"

| "\<parallel>$.\<parallel>\<^sub>E          = E($)\<^sub>d"
| "\<parallel>Upd d p v\<parallel>\<^sub>E   = \<parallel>d\<parallel>\<^sub>E[\<parallel>p\<parallel>\<^sub>E \<rightarrow> \<parallel>v\<parallel>\<^sub>E]\<^sub>?"
| "\<parallel>Del d p\<parallel>\<^sub>E     = \<parallel>d\<parallel>\<^sub>E[\<parallel>p\<parallel>\<^sub>E \<rightarrow>]\<^sub>?"

| "\<parallel>Equal a b\<parallel>\<^sub>E   = \<parallel>a\<parallel>\<^sub>E =\<^sub>? \<parallel>b\<parallel>\<^sub>E"
| "\<parallel>Less a b\<parallel>\<^sub>E    = \<parallel>a\<parallel>\<^sub>E <\<^sub>? \<parallel>b\<parallel>\<^sub>E"

| "\<parallel>L b p\<parallel>\<^sub>E       = \<parallel>p\<parallel>\<^sub>E \<longleftrightarrow>\<^sub>? b"

| "\<parallel>VCond l a b\<parallel>\<^sub>E = if\<^sub>? \<parallel>l\<parallel>\<^sub>E then \<parallel>a\<parallel>\<^sub>E else \<parallel>b\<parallel>\<^sub>E fi"
| "\<parallel>PCond l a b\<parallel>\<^sub>E = (if kindOf a = kindOf b then if\<^sub>? \<parallel>l\<parallel>\<^sub>E then \<parallel>a\<parallel>\<^sub>E else \<parallel>b\<parallel>\<^sub>E fi else \<bottom>)"
| "\<parallel>DCond l a b\<parallel>\<^sub>E = if\<^sub>? \<parallel>l\<parallel>\<^sub>E then \<parallel>a\<parallel>\<^sub>E else \<parallel>b\<parallel>\<^sub>E fi"
| "\<parallel>RCond l a b\<parallel>\<^sub>E = if\<^sub>? \<parallel>l\<parallel>\<^sub>E then \<parallel>a\<parallel>\<^sub>E else \<parallel>b\<parallel>\<^sub>E fi"

instance by intro_classes end


(* Sub-Domain Theorems *)

theorem sem_exp_cases:
  fixes ve :: valueExp and pe :: pathExp and de :: docExp and re :: pred and li :: lit
  shows sem_valueExp_cases: "\<parallel>ve\<parallel>\<^sub>E = \<bottom> \<or> (\<exists>v. \<parallel>ve\<parallel>\<^sub>E = v\<^sub>v)" 
    and sem_pathExp_cases:  "\<parallel>pe\<parallel>\<^sub>E = \<bottom> \<or> (\<exists>p. \<parallel>pe\<parallel>\<^sub>E = p\<^sub>p)"
    and sem_docExp_cases:   "\<parallel>de\<parallel>\<^sub>E = \<bottom> \<or> (\<exists>d. \<parallel>de\<parallel>\<^sub>E = d\<^sub>d)"
    and sem_pred_cases:     "\<parallel>re\<parallel>\<^sub>E = \<bottom> \<or> (\<exists>r. \<parallel>re\<parallel>\<^sub>E = r\<^sub>b)"
    and sem_lit_cases:      "\<parallel>li\<parallel>\<^sub>E = \<bottom> \<or> (\<exists>l. \<parallel>li\<parallel>\<^sub>E = l\<^sub>b)"
proof (induct ve and pe and de and re and li)
  case FVar  show ?case by simp (metis env.exhaust read_var.simps(1) read_var.simps(2)) next
  case VCond thus ?case by auto (case_tac l, auto)+ next
  case PCond thus ?case by auto (case_tac l, auto)+ next
  case DCond thus ?case by auto (case_tac l, auto)+ next
  case RCond thus ?case by auto (case_tac l, auto)+
qed auto

corollary sve [simp]:
  fixes ve :: valueExp
  shows "\<parallel>ve\<parallel>\<^sub>E \<in> valueDom"
  using sem_valueExp_cases[of ve E] by auto

corollary spe [simp]:
  fixes pe :: pathExp
  shows "\<parallel>pe\<parallel>\<^sub>E \<in> pathDom"
  using sem_pathExp_cases[of pe E] by auto

corollary sde [simp]:
  fixes de :: docExp
  shows "\<parallel>de\<parallel>\<^sub>E \<in> docDom"
  using sem_docExp_cases[of de E] by auto

corollary spr [simp]:
  fixes re :: pred
  shows "\<parallel>re\<parallel>\<^sub>E \<in> boolDom"
  using sem_pred_cases[of re E] by auto

corollary sli [simp]:
  fixes li :: lit
  shows "\<parallel>li\<parallel>\<^sub>E \<in> boolDom"
  using sem_lit_cases[of li E] by auto


lemmas sveC = sve[THEN theDomainD, simplified]
lemmas speC = spe[THEN theDomainD, simplified]
lemmas sdeC = sde[THEN theDomainD, simplified]
lemmas sprC = spr[THEN theDomainD, simplified]
lemmas sliC = sli[THEN theDomainD, simplified]

lemma orD:
  assumes "a \<or> b"
  shows "\<not> a \<Longrightarrow> b"
  using assms by simp

lemmas sxeD [dest!] = sveC[THEN orD] speC[THEN orD] sdeC[THEN orD]


theorem sli_normalize [dest!]:
  assumes "\<parallel>L (\<not> b) p\<parallel>\<^sub>E = r"
  shows "\<parallel>L b p\<parallel>\<^sub>E = (\<not>r)"
  using assms by auto


(* Finiteness Theorems *)

fun deDomain :: "docExp \<Rightarrow> bool" where
  "deDomain $.            \<longleftrightarrow> undefined"
| "deDomain (Upd d p v)   \<longleftrightarrow> deDomain d"
| "deDomain (Del d p)     \<longleftrightarrow> deDomain d"
| "deDomain (DCond b t e) \<longleftrightarrow> deDomain t \<and> deDomain e"

theorem sem_docExp_finiteO:
  fixes de :: docExp
  shows "finiteO \<parallel>de\<parallel>\<^sub>E"
  by (cases "\<parallel>de\<parallel>\<^sub>E = \<bottom>", simp) (induct de rule: deDomain.induct, auto)

corollary sem_docExp_finite [simp]:
  fixes de :: docExp assumes "\<parallel>de\<parallel>\<^sub>E = d"
  shows "finiteD d"
  using sem_docExp_finiteO[of de E] unfolding assms and finiteO.simps .


fun veDomain :: "valueExp \<Rightarrow> bool" where
  "veDomain (Val v)       \<longleftrightarrow> undefined"
| "veDomain (FVar n)      \<longleftrightarrow> undefined"
| "veDomain (PVar n)      \<longleftrightarrow> undefined"

| "veDomain (Sum m)       \<longleftrightarrow> veDomain m"
| "veDomain (Size m)      \<longleftrightarrow> veDomain m"
| "veDomain (Count m e)   \<longleftrightarrow> veDomain m \<and> veDomain e"
| "veDomain (Tally t m)   \<longleftrightarrow> veDomain m"

| "veDomain (Read d p)    \<longleftrightarrow> undefined"

| "veDomain (Scal n m)    \<longleftrightarrow> veDomain n \<and> veDomain m"
| "veDomain (VRep n m)    \<longleftrightarrow> veDomain n \<and> veDomain m"
| "veDomain (VPlus a b)   \<longleftrightarrow> veDomain a \<and> veDomain b"
| "veDomain (VMinus a b)  \<longleftrightarrow> veDomain a \<and> veDomain b"

| "veDomain (VCond b t e) \<longleftrightarrow> veDomain t \<and> veDomain e"


corollary finiteM_map_value_simple [simp, intro!]:  (* Save to do (for us), cause of the following theorem *)
  fixes m :: "value mset" assumes "finiteM m"
  shows "finiteM (mapM f m)"
  using assms by simp

corollary finiteM_map_int_simple [simp, intro!]:  (* (all value multisets from our semantics are finite) *)
  fixes m :: "int mset" assumes "finiteM m"
  shows "finiteM (mapM f m)"
  using assms by simp

corollary finiteM_asInt [simp, intro!]:
  assumes "finiteM m"
  shows "finite {e. e\<^sub>i \<in>\<^sub>* m}"
proof -
  have "finiteM (asIntM m)" using assms by (metis finiteM_map_value_simple)
  thus ?thesis by simp
qed


theorem sem_valueExp_finiteO:
  fixes ve :: valueExp
  shows "finiteO \<parallel>ve\<parallel>\<^sub>E"
  by (cases "\<parallel>ve\<parallel>\<^sub>E = \<bottom>", simp) (induct ve rule: veDomain.induct, auto)

corollary sem_valueExp_finite [simp]:
  fixes ve :: valueExp assumes "\<parallel>ve\<parallel>\<^sub>E = v\<^sub>v"
  shows "finiteM v"
  using sem_valueExp_finiteO[of ve E] unfolding assms and finiteO.simps .


(* Optimized Equations *)

corollary sem_valueExp_simps [simp]:
  shows "\<parallel>Sum m\<parallel>\<^sub>E     = sum\<^sub>? \<parallel>m\<parallel>\<^sub>E"
    and "\<parallel>Size m\<parallel>\<^sub>E    = size\<^sub>? \<parallel>m\<parallel>\<^sub>E"
    and "\<parallel>Tally t m\<parallel>\<^sub>E = tally\<^sub>? t \<parallel>m\<parallel>\<^sub>E"
  by (insert sveC[of m E]) auto

declare sem_valueExp.simps(4) [simp del]  (* Sum *)
declare sem_valueExp.simps(5) [simp del]  (* Size *)
declare sem_valueExp.simps(7) [simp del]  (* Tally *)

lemmas sem_exp_simps =
  sem_valueExp.simps(1-3) sem_valueExp_simps(1,2)
  sem_valueExp.simps(6)   sem_valueExp_simps(3)   sem_valueExp.simps(8-13)
  sem_pathExp.simps sem_docExp.simps sem_pred.simps sem_lit.simps


(* Kind Theorems *)

fun peDomain :: "pathExp \<Rightarrow> bool" where
  "peDomain Root.         \<longleftrightarrow> undefined"
| "peDomain (Path l p k)  \<longleftrightarrow> peDomain p"
| "peDomain (Kind l p)    \<longleftrightarrow> peDomain p"

| "peDomain (PRep n m)    \<longleftrightarrow> peDomain m"
| "peDomain (PPlus a b)   \<longleftrightarrow> peDomain a \<and> peDomain b"
| "peDomain (PMinus a b)  \<longleftrightarrow> peDomain a \<and> peDomain b"

| "peDomain (PCond b t e) \<longleftrightarrow> peDomain t \<and> peDomain e"

fun hasKindO :: "kind \<Rightarrow> domain \<Rightarrow> bool" ("hasKindO _ _" [1000, 999] 999) where
  "hasKindO k m\<^sub>p = hasKindM k m"
| "hasKindO k _ = True"

theorem sem_pathExp_hasKindO:
  fixes pe :: pathExp assumes "kindOf pe = k\<^sub>!"
  shows "hasKindO k \<parallel>pe\<parallel>\<^sub>E"
  using assms by (cases "\<parallel>pe\<parallel>\<^sub>E = \<bottom>", simp) (induct pe arbitrary: k rule: peDomain.induct, auto, metis+)

corollary sem_pathExp_hasKind [simp]:
  fixes pe :: pathExp assumes "kindOf pe = k\<^sub>!" and "\<parallel>pe\<parallel>\<^sub>E = p\<^sub>p"
  shows "hasKindM k p"
  using sem_pathExp_hasKindO[of pe k E] unfolding assms by simp


(* Boolean Semantics *)

class bool_sem = boolExp + sem + 
constrains x :: "'a :: {boolExp, sem}"
assumes bFalse [simp]:
  "\<parallel>False.\<parallel>\<^sub>E = False"
and bTrue [simp]:
  "\<parallel>True.\<parallel>\<^sub>E = True"
and sbe:
  "\<parallel>x\<parallel>\<^sub>E \<in> boolDom"

corollary sbeI:
  fixes x :: "'a :: bool_sem"
  assumes "\<parallel>x\<parallel>\<^sub>E = \<bottom> \<Longrightarrow> P \<bottom>" and "\<parallel>x\<parallel>\<^sub>E = True \<Longrightarrow> P True" and "\<parallel>x\<parallel>\<^sub>E = False \<Longrightarrow> P False"
  shows "P (\<parallel>x\<parallel>\<^sub>E)"
  using assms and sbe[of x E] by auto (case_tac y, auto)

corollary sbeE:
  fixes x :: "'a :: bool_sem" assumes "P (\<parallel>x\<parallel>\<^sub>E)"
    and "P \<bottom> \<Longrightarrow> \<parallel>x\<parallel>\<^sub>E = \<bottom> \<Longrightarrow> Q"
    and "P True  \<Longrightarrow> \<parallel>x\<parallel>\<^sub>E = True \<Longrightarrow> Q"
    and "P False \<Longrightarrow> \<parallel>x\<parallel>\<^sub>E = False \<Longrightarrow> Q"
  shows Q
  using assms and sbe[of x E] by auto (case_tac y, auto)

corollary sbeD [dest!]:
  fixes x :: "'a :: bool_sem" assumes "\<parallel>x\<parallel>\<^sub>E \<noteq> \<bottom>"
  shows "\<parallel>x\<parallel>\<^sub>E = True \<or> \<parallel>x\<parallel>\<^sub>E = False"
  using assms by - (rule sbeI, simp+)

corollary sbeCast [dest!]:
  fixes be :: "'a :: bool_sem" assumes "cast (\<parallel>be\<parallel>\<^sub>E) = (None :: bool option)"
  shows "\<parallel>be\<parallel>\<^sub>E = \<bottom>"
  using assms by - (rule sbeI, auto)


corollary sem_be_iff_true [simp]:
  fixes x :: "'a :: bool_sem"
  shows "\<parallel>x\<parallel>\<^sub>E \<longleftrightarrow>\<^sub>? True = \<parallel>x\<parallel>\<^sub>E"
  by (rule sbeI) auto

corollary sem_be_iff_false [simp]:
  fixes x :: "'a :: bool_sem"
  shows "\<parallel>x\<parallel>\<^sub>E \<longleftrightarrow>\<^sub>? False = \<not>\<^sub>? \<parallel>x\<parallel>\<^sub>E"
  by (rule sbeI) auto


instantiation pred and lit :: bool_sem begin instance
  by (intro_classes, unfold bexp_defs, auto simp del: bexp_normalize) end


end
