header {* Advanced Logical Syntactic Sugar *}

theory Sugar imports Syntax begin


(*------------*)
(*- Liftings -*)
(*------------*)

(* Literals *)

class lit =
fixes toLit :: "'a \<Rightarrow> lit"

class lit_sem = lit + closed +
assumes lit_sem_eq [simp]:
  "\<parallel>toLit x\<parallel>\<^sub>E = \<parallel>x\<parallel>\<^sub>E"
and lit_xprop_eq [simp]:
  "xprop p (toLit x) = xprop p x"
and lit_xcount_eq [simp]:
  "xcount c (toLit x) = xcount c x"

instantiation pred and lit :: lit_sem begin
  fun toLit_lit  :: "lit  \<Rightarrow> lit" where "toLit l = l"
  fun toLit_pred :: "pred \<Rightarrow> lit" where "toLit p = L True p"
instance by intro_classes auto end


(* Disjunctions *)

class disj =
fixes toDisj :: "'a \<Rightarrow> disj"

class disj_sem = disj + closed +
assumes disj_sem_eq [simp]:
  "\<parallel>toDisj x\<parallel>\<^sub>E = \<parallel>x\<parallel>\<^sub>E"
and disj_xprop_eq [simp]:
  "xprop p (toDisj x) = xprop p x"
and disj_xcount_eq [simp]:
  "xcount c (toDisj x) = xcount c x"

instantiation pred and lit and disj :: disj_sem begin
  fun toDisj_disj :: "disj \<Rightarrow> disj" where "toDisj d = d"
  fun toDisj_lit  :: "lit  \<Rightarrow> disj" where "toDisj l = D l DFlse"
  fun toDisj_pred :: "pred \<Rightarrow> disj" where "toDisj_pred p = toDisj (toLit p)"
instance by intro_classes auto end


(* Conjunctions *)

class conj =
fixes toConj :: "'a \<Rightarrow> conj"

class conj_sem = conj + closed +
assumes conj_sem_eq [simp]:
  "\<parallel>toConj x\<parallel>\<^sub>E = \<parallel>x\<parallel>\<^sub>E"
and conj_xprop_eq [simp]:
  "xprop p (toConj x) = xprop p x"
and conj_xcount_eq [simp]:
  "xcount c (toConj x) = xcount c x"

instantiation pred and lit and disj and conj :: conj_sem begin
  fun toConj_conj :: "conj \<Rightarrow> conj" where "toConj c = c"
  fun toConj_disj :: "disj \<Rightarrow> conj" where "toConj d = C d CTrue"
  fun toConj_lit  :: "lit  \<Rightarrow> conj" where "toConj_lit l  = toConj (toDisj l)"
  fun toConj_pred :: "pred \<Rightarrow> conj" where "toConj_pred p = toConj (toDisj p)"
instance by intro_classes auto end


(* Summary *)

corollary lift_xfold_eq:
  fixes x :: "'a :: lit_sem" and y :: "'b :: disj_sem" and z :: "'c :: conj_sem"
  shows "xprop  p (toLit  x) \<longleftrightarrow> xprop  p x"
    and "xcount c (toLit  x)  =  xcount c x"
    and "xprop  p (toDisj y) \<longleftrightarrow> xprop  p y"
    and "xcount c (toDisj y)  =  xcount c y"
    and "xprop  p (toConj z) \<longleftrightarrow> xprop  p z"
    and "xcount c (toConj z)  =  xcount c z"
  by simp+


(*---------*)
(*- Sugar -*)
(*---------*)

(* Negation *)

definition lit_negation :: "'a :: lit \<Rightarrow> lit" ("\<not>. _" [1000] 999) where
  "\<not>. x \<equiv> sNegate (toLit x)"

theorem sem_lit_negation [simp]:
  fixes x :: "'a :: {lit_sem, bool_sem}"
  shows "\<parallel>\<not>. x\<parallel>\<^sub>E = \<not>\<^sub>? \<parallel>x\<parallel>\<^sub>E"
  unfolding lit_negation_def by simp

theorem sem_liftDom_Not [simp]:
  shows "\<not>\<^sub>? \<parallel>L b p\<parallel>\<^sub>E = \<parallel>L (\<not>b) p\<parallel>\<^sub>E"
  using sprC[of p E] by auto

interpretation lit_negation_props:
  props "\<not>. x" "toLit x"
  unfolding lit_negation_def by unfold_locales


fun disj_negation :: "disj \<Rightarrow> conj" (".\<not> _" [1000] 999) where
  ".\<not> (D l d) = C (toDisj (\<not>. l)) (.\<not> d)"
| ".\<not> DFlse = CTrue"

theorem sem_disj_negation [simp]:
  shows "\<parallel>.\<not> d\<parallel>\<^sub>E = \<not>\<^sub>? \<parallel>d\<parallel>\<^sub>E"
  by (induct d, auto) (rule sbeI, auto)+

interpretation disj_negation_props:
  props ".\<not> d" d
  by (unfold_locales, auto elim!: order_trans[rotated]) (induct d, auto)+


(* Disjunctions *)

fun disj_conc :: "disj \<Rightarrow> disj \<Rightarrow> disj" where
  "disj_conc (D l e) d = sD l (disj_conc e d)"
| "disj_conc DFlse d   = d"

theorem sem_disj_conc:
  shows "\<parallel>disj_conc a b\<parallel>\<^sub>E = \<parallel>a\<parallel>\<^sub>E \<or>\<^sub>? \<parallel>b\<parallel>\<^sub>E"
  by (induct a b rule: disj_conc.induct) auto

interpretation disj_conc_props:
  props "disj_conc a b" "C a (C b CTrue)"
  by (unfold_locales, auto elim!: order_trans[rotated]) 
     (induct a b rule: disj_conc.induct, auto elim!: order_trans)+


definition disj_disj :: "'a :: disj \<Rightarrow> 'b :: disj \<Rightarrow> disj" ("_ \<or>. _" [700, 690] 690) where
  "a \<or>. b \<equiv> disj_conc (toDisj a) (toDisj b)"

abbreviation lit_imp :: "'a :: lit \<Rightarrow> 'b :: disj \<Rightarrow> disj" ("_ \<rightarrow>. _" [999, 690] 690) where
  "a \<rightarrow>. b \<equiv> \<not>. a \<or>. b"

theorem sem_disj_disj [simp]:
  fixes a :: "'a :: disj_sem" and b :: "'b :: disj_sem"
  shows "\<parallel>a \<or>. b\<parallel>\<^sub>E = \<parallel>a\<parallel>\<^sub>E \<or>\<^sub>? \<parallel>b\<parallel>\<^sub>E"
  unfolding disj_disj_def and sem_disj_conc by simp

interpretation disj_disj_props:
  props "a \<or>. b" "C (toDisj a) (C (toDisj b) CTrue)"
  unfolding disj_disj_def by unfold_locales auto


(* Conjunctions *)

fun conj_conc :: "conj \<Rightarrow> conj \<Rightarrow> conj" where
  "conj_conc (C d e) c = sC d (conj_conc e c)"
| "conj_conc CTrue   c = c"

theorem sem_conj_conc:
  shows "\<parallel>conj_conc d c\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E \<and>\<^sub>? \<parallel>c\<parallel>\<^sub>E"
  by (induct d c rule: conj_conc.induct) auto


datatype cc = CC conj conj

instantiation cc :: xfold begin
  fun xprop_cc  where "xprop  f (CC a b) \<longleftrightarrow> xprop  f a \<and> xprop  f b"
  fun xcount_cc where "xcount f (CC a b)  =  xcount f a + xcount f b"
  fun xvars_cc  where "xvars  f (CC a b)  =  xvars  f a \<union> xvars  f b"
  fun xlits_cc  where "xlits  f (CC a b)  =  xlits  f a @ xlits  f b"
  fun xp_cc :: "bool     xfold \<Rightarrow> cc \<Rightarrow> bool"     where "xp_cc _ _ = True"
  fun xc_cc :: "nat      xfold \<Rightarrow> cc \<Rightarrow> nat"      where "xc_cc _ _ = 0"
  fun xv_cc :: "name set xfold \<Rightarrow> cc \<Rightarrow> name set" where "xv_cc _ _ = {}"
  fun xl_cc :: "lit list xfold \<Rightarrow> cc \<Rightarrow> lit list" where "xl_cc _ _ = []"
instance by intro_classes (case_tac x, clarsimp, ((subst xcount_to_xprop, simp+)+)?)+ end

instantiation cc :: xmap begin
  fun xmap_cc where "xmap f (CC a b) = CC (xmap f a) (xmap f b)"
  fun xm_cc :: "xmap \<Rightarrow> cc \<Rightarrow> cc" where "xm_cc _ x = x"
instance by intro_classes (case_tac x, simp) end

interpretation conj_conc_props:
  props "conj_conc a b" "CC a b"
  by (unfold_locales, auto elim!: order_trans[rotated]) 
     (induct a b rule: conj_conc.induct, auto elim!: order_trans)+


definition conj_conj :: "'a :: conj \<Rightarrow> 'b :: conj \<Rightarrow> conj" ("_ \<and>. _" [700, 680] 680) where
  "a \<and>. b \<equiv> conj_conc (toConj a) (toConj b)"

theorem sem_conj_conj [simp]:
  fixes a :: "'a :: conj_sem" and b :: "'b :: conj_sem"
  shows "\<parallel>a \<and>. b\<parallel>\<^sub>E = \<parallel>a\<parallel>\<^sub>E \<and>\<^sub>? \<parallel>b\<parallel>\<^sub>E"
  unfolding conj_conj_def and sem_conj_conc by simp

theorem sem_cnf_conj_conj_closed:
  fixes a :: "'a :: conj_sem" and b :: "'b :: conj_sem"
  shows "\<parallel>\<forall>. (a \<and>. b)\<parallel>\<^sub>R\<^sub>\<up> = \<parallel>\<forall>. toConj a\<parallel>\<^sub>R \<and>\<^sub>? \<parallel>\<forall>. toConj b\<parallel>\<^sub>R"
proof -
  { fix f assume "\<parallel>a\<parallel>\<^sub>R\<^sub>+\<^sub>f = False \<or> \<parallel>b\<parallel>\<^sub>R\<^sub>+\<^sub>f = False" hence "\<exists>f. \<parallel>a\<parallel>\<^sub>R\<^sub>+\<^sub>f \<and>\<^sub>? \<parallel>b\<parallel>\<^sub>R\<^sub>+\<^sub>f = False"
    by (rule_tac x=f in exI) auto } 
  thus ?thesis unfolding sem_cnf.simps by auto
qed

corollary sem_cnf_conj_conj [simp]:
  fixes a :: "'a :: conj_sem" and b :: "'b :: conj_sem"
  shows "\<parallel>\<forall>. (a \<and>. b)\<parallel>\<^sub>E = \<parallel>\<forall>. toConj a\<parallel>\<^sub>E \<and>\<^sub>? \<parallel>\<forall>. toConj b\<parallel>\<^sub>E"
  using sem_cnf_conj_conj_closed by (cases E) auto

interpretation conj_conj_props:
  props "a \<and>. b" "CC (toConj a) (toConj b)"
  unfolding conj_conj_def by unfold_locales auto


(* Disjunction with Conjunctions *)

fun disj_conc_conj :: "disj \<Rightarrow> conj \<Rightarrow> conj" where
  "disj_conc_conj d CTrue = CTrue"
| "disj_conc_conj d (C e c) = (d \<or>. e) \<and>. (disj_conc_conj d c)"

theorem sem_disj_conc_conj:
  shows "\<parallel>disj_conc_conj d c\<parallel>\<^sub>E = \<parallel>d\<parallel>\<^sub>E \<or>\<^sub>? \<parallel>c\<parallel>\<^sub>E"
  by (induct d c rule: disj_conc_conj.induct, auto) (rule sbeI, auto)+

interpretation disj_conc_conj_props:
  props_eq "disj_conc_conj d c" "C d c"
  by unfold_locales (induct d c rule: disj_conc_conj.induct, auto)+


definition disj_disj_conj :: "'a :: disj \<Rightarrow> conj \<Rightarrow> conj" ("_ .\<or> _" [700, 690] 680) where
  "a .\<or> b \<equiv> disj_conc_conj (toDisj a) b"

abbreviation lit_imp_conj :: "'a :: lit \<Rightarrow> conj \<Rightarrow> conj" ("_ \<longrightarrow>. _" [700, 700] 700) where
  "a \<longrightarrow>. b \<equiv> \<not>. a .\<or> b"


theorem sem_disj_disj_conj [simp]:
  fixes a :: "'a :: disj_sem"
  shows "\<parallel>a .\<or> b\<parallel>\<^sub>E = \<parallel>a\<parallel>\<^sub>E \<or>\<^sub>? \<parallel>b\<parallel>\<^sub>E"
  using sem_disj_conc_conj unfolding disj_disj_conj_def by simp

theorem sem_disj_disj_conj_closed:
  fixes a :: "'a :: {lit_sem, bool_sem}" assumes "closed a"
  shows "\<parallel>\<forall>. (a \<longrightarrow>. b)\<parallel>\<^sub>R\<^sub>\<up> = \<not>\<^sub>? \<parallel>a\<parallel>\<^sub>R \<or>\<^sub>? \<parallel>\<forall>. b\<parallel>\<^sub>R"
  using assms unfolding sem_cnf.simps by auto (rule sbeI, auto)+

corollary sem_disj_disj_conj_closed_general [simp]:
  fixes a :: "'a :: {lit_sem, bool_sem}" assumes "closed a"
  shows "\<parallel>\<forall>. (a \<longrightarrow>. b)\<parallel>\<^sub>E = \<not>\<^sub>? \<parallel>a\<parallel>\<^sub>E \<or>\<^sub>? \<parallel>\<forall>. toConj b\<parallel>\<^sub>E"
  using assms and sem_disj_disj_conj_closed by (cases E) auto

interpretation disj_disj_conj_props:
  props_eq "a .\<or> b" "C (toDisj a) (toConj b)"
  unfolding disj_disj_conj_def by (unfold_locales) simp+


(* Higher Sugar Predicates *)

abbreviation unique_key :: "valueExp \<Rightarrow> pred" ("unique.key _" [1000] 999) where
  "unique.key v \<equiv> tally. TKey v =. 1."

fun disj_proper :: "pathExp \<Rightarrow> disj" (".proper _" [1000] 999) where
  ".proper Root.        = False."
| ".proper (Path l p k) = unique.key k \<rightarrow>. .proper p"
| ".proper (Kind l p)   = True."
| ".proper _            = undefined"  (* we define the function for basic paths only *)

abbreviation conj_unique :: "pathExp \<Rightarrow> conj" (".unique _" [1000] 999) where
  ".unique p \<equiv> .\<not> (.proper p)"

abbreviation conj_in_dom :: "pathExp \<Rightarrow> conj" ("_ \<in>\<^sub>1.$" [700] 700) where
  "p \<in>\<^sub>1.$ \<equiv> .unique p \<and>. unique. $.(p)."

abbreviation conj_not_in_dom :: "pathExp \<Rightarrow> conj" ("_ \<notin>\<^sub>1.$" [700] 700) where
  "p \<notin>\<^sub>1.$ \<equiv> .unique p \<and>. empty. $.(p)."


theorem sem_disj_proper_bot [dest]:
  assumes "\<parallel>.proper p\<parallel>\<^sub>E = \<bottom>" and "basic p"
  shows "\<parallel>p\<parallel>\<^sub>E = \<bottom>"
  using assms and sTally_sem_simplified  by (induct p rule: disj_proper.induct) auto

theorem sem_disj_proper [simp]:
  assumes "basic p" and "\<parallel>p\<parallel>\<^sub>E = q\<^sub>p"
  shows "\<parallel>.proper p\<parallel>\<^sub>E = proper q"
  using assms using sTally_sem_simplified sEqual_sem_simplified
  by (induct p arbitrary: q rule: disj_proper.induct) (auto, drule arg_cong[of _ _ finiteM], auto)


theorem basic_disj_proper [simp]:
  assumes "basic p"
  shows "strict (.proper p)" and "straight (.proper p)"
  using assms by (induct rule: disj_proper.induct) simp+

theorem closed_disj_proper [simp]:
  assumes "basic p" and "closed p"
  shows "closed (.proper p)"
  using assms by (induct rule: disj_proper.induct) simp+

theorem wform_disj_proper [simp]:
  assumes "basic p" and "wform p"
  shows "wform (.proper p)"
  using assms by (induct p rule: disj_proper.induct) simp+

end

